import json
import os
import shutil
import subprocess
import sys

from PyInquirer import prompt
from github import Github, GithubException, UnknownObjectException
from src.cobra.conf.config import Config
from src.cobra.utils.constants import Texts, Structures, Paths


def get_object_from_list_by_name(filter_str, input_list):
    return next(filter(lambda x: filter_str in x.name, input_list), None)


def build_question(interface_type: str, message: str, name: str, iter_list: list):
    question = {
        'type': interface_type,
        'name': name,
        'message': message,
    }

    if interface_type == 'list':
        if type(iter_list[0]) == str:
            choices = iter_list
        else:
            choices = [item.name for item in iter_list]
        question['choices'] = choices

    return question


def ask_questions(types: list, messages: list, names: list, iter_list: list):
    questions = [build_question(t, m, n, l) for t, m, n, l in zip(types, messages, names, iter_list)]
    return prompt(questions)


def check_env_exists(name):
    # execute command
    envs = subprocess.check_output(['conda', 'env', 'list', '--json']).decode('utf-8')
    # create a dict
    envs = json.loads(envs.replace('\n', ''))['envs']
    # just get the basename of the conda env path
    envs = [os.path.basename(e) for e in envs]
    if name in envs:
        return True
    return False


def dialog_username_password():
    # username and password
    username_password = ask_questions(['input', 'password'],
                                      [Texts.USERNAME_TEXT, Texts.PASSWORD_TEXT],
                                      ['username', 'password'],
                                      [[], []])
    username = username_password['username']
    password = username_password['password']
    github = Github(login_or_token=username, password=password)
    return github, username, password


def user_password_dialog(error=None):
    if error is None:
        error = {}
    try:
        if not error:
            github, username, password = dialog_username_password()
            auth = github.get_user().create_authorization(scopes=Structures.AUTH_SCOPES,
                                                          note='cobra')
        else:
            if error['key'] == 0:
                print(error['message'])
                github, username, password = dialog_username_password()
                auth = github.get_user().create_authorization(scopes=Structures.AUTH_SCOPES,
                                                              note='cobra')

            if error['key'] == 1:
                print(error['message'])
                github, username, password = error['cred']
                tfa = ask_questions(['input'], [Texts.TFA_TEXT], ['tfa'], [[]])['tfa']
                auth = github.get_user().create_authorization(scopes=Structures.AUTH_SCOPES,
                                                              onetime_password=tfa,
                                                              note='cobra')

        sec = auth.token or ""
        github = Github(sec)

        # to test if the github object has a correct authentication
        github.get_user().id

        Config.getInstance().username = username
        Config.getInstance().sec = sec
    except GithubException as e:
        error_msg = {}
        if e.status == 422:
            error_msg['message'] = Texts.TOKEN_ALREADY_EXISTS_TEXT
            error_msg['key'] = 1
        if e.status == 401:
            if e.data['message'] == 'Bad credentials':
                error_msg['key'] = 0
            else:
                error_msg['key'] = 1
            error_msg['message'] = e.data['message']
        error_msg['cred'] = (github, username, password)
        return error_msg


def duplication_check(project):
    # [repo, pull, conda]
    res = []
    None if check_repo_exists('/'.join([project.repo_user, project.repo_name])) else res.append('create_repo')
    None if check_git_pull() else res.append('pull_repo')
    None if check_env_exists(project.conda_name) else res.append('create_environment')
    return res


def check_git_pull():
    return os.path.exists(os.path.join(os.getcwd(), '.git'))


def check_repo_exists(full_repo_name):
    try:
        Github().get_repo(full_repo_name)
    except UnknownObjectException:
        return False
    except GithubException:
        return False
    return True


def create_files_folders():
    # create folder
    os.mkdir(Paths.COBRA_FOLDER)

    # create conf file
    open(Paths.CONF_FILE_PATH, 'a').close()
    # Config.getInstance().write_into_yaml_file(Constants.CONF_FILE_PATH, **Constants.CONF_DICT)
    # project file {'TestProject': {infos}}
    open(Paths.PROJECT_FILE_PATH, 'a').close()


def check_requirements():
    if shutil.which("conda") is None:
        raise Exception(
            "Conda is not installed! https://docs.conda.io/projects/conda/en/latest/user-guide/install/")
    if sys.version_info[0] < 3 and sys.version_info[1] < 5:
        raise Exception("The default python version is lower then 3.5. Please update!")


def first_usage():
    """
    Checks if the pocr config file is existing.
    If this file is existing we know that it is not first usage and so the function returns false.
    Else vise versa.

    :return:
        boolean: if its first usage or not
    """
    return not os.path.exists(Paths.COBRA_FOLDER)


def get_github_user():
    # get github user
    github = Github(Config.getInstance().sec)
    user = github.get_user()
    return user


def delete_path(path: str):
    # check if path is pointing to a file or folder
    if os.path.exists(path):
        try:
            if os.path.isfile(path):
                os.remove(path)
            else:
                shutil.rmtree(path)
        except PermissionError:
            print("Permission error for deleting the folder."
                  " Please delete it by hand or try again.")



import os

from src.cobra.conf.config import Config
from src.cobra.connenction_types import ConnectionType
from src.cobra.utils.constants import Paths


class TestConfig:
    def test_get_instance(self):
        config1 = Config.getInstance()
        config2 = Config.getInstance()
        assert config1 == config2

    def test_load_config(self, tmp_path, monkeypatch):
        config_content = """!Config
                            _connection_type: &id001 !ConnectionType
                              _name: ssh
                              _url: 'git@github.com:'
                            _used_vcs: !VCS
                              _connection_types:
                              - !ConnectionType
                                _name: https
                                _url: https://github.com/
                              - *id001
                              _name: Github
                            _username: testUser"""

        conf_file_path = tmp_path / "load_config"
        conf_file_path.write_text(config_content)

        monkeypatch.setattr(Paths, "CONF_FILE_PATH", conf_file_path.__str__())

        conf = Config.getInstance()

        conf.load_config()
        conf._sec = "testSec"
        assert conf.used_vcs.name == "Github"
        assert conf.connection_type.name == "ssh"
        assert conf.connection_type.url == "git@github.com:"
        assert conf.username == "testUser"
        assert conf._sec == "testSec"
        assert conf.vcses[0].name == "Github"
        conf.used_vcs = None

    def test_save_config(self, tmp_path, monkeypatch):
        conf = Config.getInstance()

        conf_file_path = tmp_path / "save_config"

        conf.username = "Test"
        conf.connection_type = ConnectionType("TestConType", "TestUrl")

        monkeypatch.setattr(Paths, "CONF_FILE_PATH", conf_file_path.__str__())

        conf.save_config()

        assert os.path.exists(conf_file_path.__str__())

        import yaml
        with open(conf_file_path.__str__(), 'r') as f:
            config = yaml.load(f, Loader=yaml.Loader)
            assert config.username == "Test"
            assert config.connection_type.name == "TestConType"
            assert config.connection_type.url == "TestUrl"
            assert config.used_vcs is None



from src.cobra.utils.utils import first_usage, check_git_pull, create_files_folders, delete_path
from src.cobra.utils.constants import Paths


class TestUtils:
    def test_not_first_usage(self, tmp_path, monkeypatch):
        cobra_home_dir = tmp_path / ".cobra"
        cobra_home_dir.mkdir()
        monkeypatch.setattr(Paths, "COBRA_FOLDER", cobra_home_dir)
        assert not first_usage()

    def test_first_usage(self, tmp_path, monkeypatch):
        cobra_home_dir = tmp_path / ".cobra"
        monkeypatch.setattr(Paths, "COBRA_FOLDER", cobra_home_dir)
        assert first_usage()

    def test_check_git_pull(self, tmp_path, monkeypatch):
        git_home_dir = tmp_path / ".git"
        git_home_dir.mkdir()

        monkeypatch.setattr("os.getcwd", lambda: tmp_path.__str__())
        assert check_git_pull()

    def test_check_git_not_pull(self, tmp_path, monkeypatch):
        monkeypatch.setattr("os.getcwd", lambda: tmp_path.__str__())
        assert not check_git_pull()

    def test_create_files_folders(self, tmp_path, monkeypatch):
        import os
        cobra_home_dir = tmp_path / ".cobra"
        monkeypatch.setattr(Paths, "COBRA_FOLDER", cobra_home_dir)
        monkeypatch.setattr(Paths, "CONF_FILE_PATH", cobra_home_dir / "config")
        monkeypatch.setattr(Paths, "PROJECT_FILE_PATH", cobra_home_dir / "projects")
        create_files_folders()

        cobra_path = (tmp_path / ".cobra").__str__()
        # checks folder exists
        assert os.path.exists(cobra_path)
        # checks project files exists
        assert os.path.exists(os.path.join(cobra_path, "projects"))
        # checks config files exists
        assert os.path.exists(os.path.join(cobra_path, "config"))

    def test_delete_path_folder(self, tmp_path):
        import os
        dummy_folder = tmp_path / "dummy"
        dummy_folder.mkdir()
        assert os.path.exists(dummy_folder.__str__())
        delete_path(dummy_folder.__str__())
        assert not os.path.exists(dummy_folder.__str__())

    def test_delete_path_file(self, tmp_path):
        import os
        dummy_file = tmp_path / "dummy.txt"
        dummy_file.mkdir()
        assert os.path.exists(dummy_file.__str__())
        delete_path(dummy_file.__str__())
        assert not os.path.exists(dummy_file.__str__())



import yaml

from src.cobra.connenction_types import ConnectionType


class VCS(yaml.YAMLObject):
    yaml_tag = u'!VCS'

    def __init__(self, name, connection_types):
        self.name = name
        self.connection_types = self.__init_connection_types(connection_types)

    def __repr__(self):
        return "{}".format(self.name)

    # GETTERS / SETTERS

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        self._name = value

    @property
    def connection_types(self):
        return self._connection_types

    @connection_types.setter
    def connection_types(self, value):
        self._connection_types = value

    @staticmethod
    def __init_connection_types(connection_types):
        con_type_list = []
        for name, url in connection_types.items():
            con_type_list.append(ConnectionType(name, url))
        return con_type_list



import argparse
import sys


def get_params():
    parser = argparse.ArgumentParser(description="Cobra command line tool")
    parser.add_argument("--install", help="Start the installation wizard.", action="store_true")
    parser.add_argument("--test", help="Puts cobra into testing mode", action="store_true")
    parser.add_argument("--clear", help="Clears all user files", action="store_true")
    # TODO option to create the repo private

    subparsers = parser.add_subparsers(help="The different functions of cobra", dest='command')

    # create subcommand
    create_parser = subparsers.add_parser("create", help="Create a new cobra project")
    create_group = create_parser.add_mutually_exclusive_group(required=True)
    create_group.add_argument('-n', '--name',
                              help="Name of the project",
                              type=str)
    create_group.add_argument('-f', '--from-file',
                              help="Searches in the current directory for a .cobra file"
                                   " and creates a project based on this",
                              action="store_true")

    create_parser.add_argument('-p', '--python-version',
                               help="Python version for the project. Default: 3.5",
                               required=False,
                               type=float,
                               default=3.5)
    create_parser.add_argument('-r', '--repo-name',
                               help="If you already have a github repo for the project."
                                    " Enter it in the following format: (username/reponame)",
                               required=False,
                               type=str)
    create_parser.add_argument('-c', '--conda-name',
                               help="If you already have a conda environment for the project. Enter the name.",
                               required=False,
                               type=str)
    create_parser.add_argument('-gh', '--git-hook',
                               help="Does not install a post-commit git hook.",
                               action="store_false")
    # list subcommand
    list_parser = subparsers.add_parser("list", help="Lists all existing cobra projects")

    # # update subcommand
    # update_parser = subparsers.add_parser('update', help="Updates a existing cobra project")

    # remove subcommand
    remove_parser = subparsers.add_parser('remove', help="Remove a cobra project")
    remove_parser.add_argument('-n', '--name',
                               help="Name of the cobra project to remove",
                               required=True,
                               type=str)
    remove_parser.add_argument('-r', '--repo',
                               help="Also removes the remote repository",
                               action="store_true")
    remove_parser.add_argument('-c', '--conda',
                               help="Also removes the conda environment from the system",
                               action="store_true")
    remove_parser.add_argument('-f', '--folder',
                               help="Also removes the local folder",
                               action="store_true")
    remove_parser.add_argument('-a', '--remove-all',
                               help="Remove everything (repo, folder, conda)",
                               action="store_true")

    # config subcommand TODO Do we need that option?

    if len(sys.argv) == 1:
        parser.print_help()
        sys.exit(1)

    return parser.parse_args()



import copy
import yaml
import keyring
import pkg_resources

from src.cobra.utils.constants import Paths
from src.cobra.vcs import VCS


class Config(yaml.YAMLObject):
    __instance = None
    yaml_tag = u'!Config'

    def __init__(self, _used_vcs=None, _connection_type=None, _username=''):
        if Config.__instance:
            self.getInstance()
        self.vcses = []
        self.used_vcs = _used_vcs
        self.connection_type = _connection_type
        self.user_password_domain = "cobra_user_sec"
        self.username = _username
        self.sec = ''

    # GETTER / SETTER
    @property
    def vcses(self):
        if not self._vcses:
            self.__load_vcs()
        return self._vcses

    @vcses.setter
    def vcses(self, value):
        self._vcses = value

    @property
    def used_vcs(self):
        return self._used_vcs

    @used_vcs.setter
    def used_vcs(self, value):
        self._used_vcs = value

    @property
    def connection_type(self):
        return self._connection_type

    @connection_type.setter
    def connection_type(self, value):
        self._connection_type = value

    @property
    def username(self):
        return self._username

    @username.setter
    def username(self, value):
        self._username = value

    @property
    def sec(self):
        return self._sec

    @sec.setter
    def sec(self, value):
        self._sec = value

    # STATIC / CLASS
    @classmethod
    def getInstance(cls):
        if cls.__instance is None:
            cls.__instance = Config()
        return cls.__instance

    # PRIVATE
    def __load_vcs(self):
        vcs = pkg_resources.resource_string(__name__, Paths.PACKAGE_VCS_PATH)
        yaml_file = yaml.safe_load_all(vcs.decode('UTF-8'))
        for vcs_list in yaml_file:
            for k, v in vcs_list.items():
                self._vcses.append(VCS(k, v['connection_types']))

    def __load_user_cred(self):
        self._sec = keyring.get_password(self.user_password_domain, self.username)

    # PUBLIC
    def save_user_cred(self):
        keyring.set_password(self.user_password_domain, self.username, self.sec)

    def load_config(self):
        with open(Paths.CONF_FILE_PATH, 'r') as f:
            config = yaml.load(f, Loader=yaml.Loader)
            self.connection_type = config.connection_type
            self.used_vcs = config.used_vcs
            self.username = config.username

            self.__load_vcs()
            self.__load_user_cred()

    def save_config(self):
        self.save_user_cred()
        with open(Paths.CONF_FILE_PATH, 'w') as f:
            obj_copy = copy.deepcopy(self.getInstance())
            del obj_copy._sec
            del obj_copy._vcses
            del obj_copy.user_password_domain
            yaml.dump(obj_copy, f)


if __name__ == '__main__':
    t = Config.getInstance()
    t.write_into_config(**{'test': 'test'})



import copy
import inspect
import os
import shutil

import pkg_resources
import yaml
from src.cobra.utils.constants import Paths
from src.cobra.utils.exceptions import ProjectNameAlreadyExists, NoCobraFileFound
from src.cobra.utils.module_functions import ModuleFunctions
from src.cobra.utils.utils import duplication_check


class Project(yaml.YAMLObject):
    yaml_tag = u'!Project'

    def __init__(self, project_path, project_name, conda_name, repo_name, repo_user, vcs, python_version):
        self.project_path = os.path.join(project_path, project_name)
        self.project_name = project_name
        self.conda_name = conda_name
        self.repo_name = repo_name
        self.repo_user = repo_user
        self.vcs = vcs
        self.python_version = python_version

    def __repr__(self):
        return "{}\t{}\t{}\t{}".format(self._project_name, self._conda_name, self._vcs, self._python_version)

    def create_project_file(self):
        with open(os.path.join(self.project_path, ".cobra"), 'w') as f:
            obj_copy = copy.deepcopy(self)
            del obj_copy._project_path
            yaml.dump(obj_copy, f)

    def append_project(self):
        yaml_dict = Project.project_exists(self.project_name)
        yaml_dict[self.project_name] = self
        with open(Paths.PROJECT_FILE_PATH, 'a') as f:
            yaml.dump(yaml_dict, f)

    # STATIC
    @staticmethod
    def remove_project(project_name):
        with open(Paths.PROJECT_FILE_PATH, 'r') as f:
            yaml_dict = yaml.load(f, Loader=yaml.Loader) or {}
        project = yaml_dict[project_name]
        del yaml_dict[project_name]

        if yaml_dict:
            with open(Paths.PROJECT_FILE_PATH, 'w') as f:
                yaml.dump(yaml_dict, f)
        else:
            open(Paths.PROJECT_FILE_PATH, 'w').close()

        return project

    @staticmethod
    def get_projects():
        with open(Paths.PROJECT_FILE_PATH, 'r') as f:
            return yaml.load(f, Loader=yaml.Loader) or None

    @staticmethod
    def project_exists(project_name: str):
        with open(Paths.PROJECT_FILE_PATH, 'r') as f:
            yaml_dict = yaml.load(f, Loader=yaml.Loader)
        if type(yaml_dict) is not dict:
            return {}
        if project_name in yaml_dict:
            raise ProjectNameAlreadyExists()
        return yaml_dict

    @staticmethod
    def project_from_file():
        # search for the .cobra file
        if os.path.exists(".cobra"):
            with open(".cobra", 'r') as f:
                project = yaml.load(f, Loader=yaml.Loader)
        else:
            raise NoCobraFileFound()
        return project

    # GETTERS / SETTERS

    @property
    def project_path(self):
        return self._project_path

    @project_path.setter
    def project_path(self, value):
        self._project_path = value

    @property
    def project_name(self):
        return self._project_name

    @project_name.setter
    def project_name(self, value):
        self._project_name = value

    @property
    def conda_name(self):
        return self._conda_name

    @conda_name.setter
    def conda_name(self, value):
        self._conda_name = value

    @property
    def vcs(self):
        return self._vcs

    @vcs.setter
    def vcs(self, value):
        self._vcs = value

    @property
    def python_version(self):
        return self._python_version

    @python_version.setter
    def python_version(self, value):
        self._python_version = value

    @property
    def repo_name(self):
        return self._repo_name

    @repo_name.setter
    def repo_name(self, value):
        self._repo_name = value

    @property
    def repo_user(self):
        return self._repo_user

    @repo_user.setter
    def repo_user(self, value):
        self._repo_user = value

    @staticmethod
    def create_project_parts(project, git_hook, **kwargs):
        # check for modules existing
        check_mask = duplication_check(project)
        MODULE_FUNCTIONS = dict(inspect.getmembers(ModuleFunctions, predicate=inspect.isfunction))

        for mask in check_mask:
            MODULE_FUNCTIONS[mask](project)

        if git_hook:
            copy_to = os.path.join(project.project_path, '.git', 'hooks', 'post-commit')
            shutil.copy(pkg_resources.resource_filename(__name__, Paths.PACKAGE_GIT_HOOK_PATH), copy_to)



from src.cobra.project import Project


def create_mock_cobra_file(path):
    content = """!Project
                _conda_name: condaName
                _project_name: projectName
                _python_version: 3.6
                _repo_name: repoName
                _repo_user: repoUser
                _vcs: null"""

    project_file_path = path / ".cobra"
    project_file_path.write_text(content)


def create_mock_project(path):
    project_path = path
    project_name = "projectName"
    conda_name = "condaName"
    repo_name = "repoName"
    repo_user = "repoUser"
    vcs = None
    python_version = 3.6

    return Project(project_path.__str__(), project_name, conda_name, repo_name, repo_user, vcs, python_version)



import os
from pathlib import Path

import pytest
from src.cobra.project import Project
from src.cobra.utils.constants import Paths
from src.cobra.utils.exceptions import ProjectNameAlreadyExists

from tests.utils.utils import create_mock_cobra_file, create_mock_project


class TestProject:
    def test_create_project_file(self, tmp_path):
        project = create_mock_project(tmp_path)

        # create project folder
        complete_path = Path(project.project_path)
        complete_path.mkdir()

        project.create_project_file()
        assert os.path.exists((complete_path / '.cobra').__str__())
        assert os.stat((complete_path / '.cobra').__str__()).st_size != 0

    def test_project_from_file(self, tmp_path, monkeypatch):
        create_mock_cobra_file(tmp_path)

        monkeypatch.chdir(tmp_path)
        project = Project.project_from_file()
        project.project_path = (tmp_path / project.project_name).__str__()

        assert project.project_path == (tmp_path / 'projectName').__str__()
        assert project.project_name == "projectName"
        assert project.python_version == 3.6
        assert project.repo_name == "repoName"
        assert project.repo_user == "repoUser"
        assert project.vcs is None
        assert project.__repr__() == "{}\t{}\t{}\t{}".format("projectName", 'condaName', None, 3.6)

    def test_append_project(self, tmp_path, monkeypatch):
        project_file_path = tmp_path / "project"
        monkeypatch.setattr(Paths, 'PROJECT_FILE_PATH', project_file_path.__str__())
        project_file_path.write_text("")

        project = create_mock_project(tmp_path)
        project.append_project()

        with pytest.raises(ProjectNameAlreadyExists) as exp:
            Project.project_exists(project.project_name)

        assert os.stat(project_file_path.__str__()).st_size != 0
        assert exp.type is ProjectNameAlreadyExists

    def test_remove_project(self, tmp_path, monkeypatch):
        project_file_path = tmp_path / "project"
        monkeypatch.setattr(Paths, 'PROJECT_FILE_PATH', project_file_path.__str__())
        project_file_path.write_text("")
        create_mock_project(project_file_path).append_project()

        project = Project.remove_project("projectName")

        assert os.stat(project_file_path.__str__()).st_size == 0
        assert type(Project.project_exists(project.project_name)) == dict

    def test_get_projects_empty(self, tmp_path, monkeypatch):
        project_file_path = tmp_path / "project"
        monkeypatch.setattr(Paths, 'PROJECT_FILE_PATH', project_file_path.__str__())
        project_file_path.write_text("")

        assert Project.get_projects() is None

    def test_get_projects(self, tmp_path, monkeypatch):
        project_file_path = tmp_path / "project"
        monkeypatch.setattr(Paths, 'PROJECT_FILE_PATH', project_file_path.__str__())
        project_file_path.write_text("")
        create_mock_project(project_file_path).append_project()

        projects = Project.get_projects()
        project = projects['projectName']

        assert projects is not None
        assert project.project_name == 'projectName'
        assert project.conda_name == 'condaName'
        assert project.project_path == (project_file_path / 'projectName').__str__()
        assert project.python_version == 3.6
        assert project.repo_name == 'repoName'
        assert project.repo_user == 'repoUser'
        assert project.vcs is None

    def test_project_exists_not(self, tmp_path, monkeypatch):
        project_file_path = tmp_path / "project"
        monkeypatch.setattr(Paths, 'PROJECT_FILE_PATH', project_file_path.__str__())
        project_file_path.write_text("")

        assert os.stat(project_file_path.__str__()).st_size == 0
        assert Project.project_exists("projectName") == {}

    def test_project_exists(self, tmp_path, monkeypatch):
        project_file_path = tmp_path / "project"
        monkeypatch.setattr(Paths, 'PROJECT_FILE_PATH', project_file_path.__str__())
        project_file_path.write_text("")
        create_mock_project(project_file_path).append_project()

        with pytest.raises(ProjectNameAlreadyExists) as exp:
            print(Project.project_exists("projectName"))

        assert os.stat(project_file_path.__str__()).st_size != 0
        assert exp.type == ProjectNameAlreadyExists



import os
import subprocess
import sys

from github import UnknownObjectException
from tabulate import tabulate

from src.cobra.conf.config import Config
from src.cobra.project import Project
from src.cobra.utils.command_line import get_params
from src.cobra.utils.constants import Texts
from src.cobra.utils.exceptions import ProjectNameAlreadyExists
from src.cobra.utils.utils import get_object_from_list_by_name, ask_questions, user_password_dialog, \
    check_requirements, first_usage, get_github_user, check_env_exists, \
    delete_path, create_files_folders


def main():
    # read args
    args = get_params()

    # requirements check
    check_requirements()

    if args.install:
        if first_usage():
            installation()
            print('cobra successfully installed!')
            sys.exit(1)
        else:
            print("cobra is already installed!")
            sys.exit(1)

    if first_usage():
        installation()
        print('cobra successfully installed!')
        sys.exit(1)

    if args.command == 'create':
        create(**args.__dict__)

    if args.command == 'list':
        listing(**args.__dict__)

    if args.command == 'remove':
        remove(**args.__dict__)

    if args.clear:
        subprocess.run("./cobra/clean.sh", shell=True)


def installation():
    """
    With this method the user installs the tool. That means he creates the config and the project file,
    enters the user credentials as well as defines the connection type.
    """

    # ask for the vcs which are stored in the vc.yml file
    vcs_selection = ask_questions(['list'], [Texts.VCS_SELECT_TEXT], ['vcs'], [Config.getInstance().vcses])
    used_vcs = Config.getInstance().used_vcs = get_object_from_list_by_name(vcs_selection['vcs'],
                                                                            Config.getInstance().vcses)

    # finish vcs install (ssh http selection; username and password or token)
    error = user_password_dialog()
    while error:
        error = user_password_dialog(error) or {}

    con_selection = ask_questions(['list'], [Texts.CON_SELECT_TEXT], ['con_type'], [used_vcs.connection_types])
    Config.getInstance().connection_type = get_object_from_list_by_name(con_selection['con_type'],
                                                                        used_vcs.connection_types)

    # create files and folders
    create_files_folders()

    # save infos
    Config.getInstance().save_config()


def create(name, python_version, from_file, **kwargs):
    # load config
    Config.getInstance().load_config()

    if from_file:
        print("Creating project from file")
        project = Project.project_from_file()
        print(".cobra file found. Continue processing...")
        name = project.project_name

        if os.path.basename(os.getcwd()) != project.repo_name:
            delete_path(os.path.join(os.getcwd(), '.cobra'))
            project.project_path = os.path.join(os.getcwd(), project.repo_name)
        else:
            project.project_path = os.getcwd()
    else:
        project = Project(os.getcwd(),
                          name,
                          "repo_name" in kwargs if kwargs['repo_name'] else name,
                          "conda_name" in kwargs if kwargs['conda_name'] else name,
                          Config.getInstance().username,
                          Config.getInstance().used_vcs,
                          python_version)

    # check for if project exists
    Project.project_exists(name)

    # creat missing elements
    Project.create_project_parts(project, **kwargs)

    project.create_project_file()

    # save path, conda name, name, git link, python version into project file
    project.append_project()

    print("****************************************************")
    print("****************************************************")
    print("Successfully created project {}".format(project.project_name))
    print("****************************************************")
    print("****************************************************")


def listing(**kwargs):
    projects = Project.get_projects()
    if projects:
        headers = vars(next(iter(projects.values()))).keys()
        headers = [h[1:] for h in headers]
        projects = [list(vars(p).values()) for p in projects.values()]
        print(tabulate(projects, headers=headers))
    else:
        print("There are no cobra projects! Use the 'create' command to create some projects.")


def remove(name, folder, repo, conda, remove_all, **kwargs):
    # remove the project from the project file
    try:
        Project.project_exists(name)
        print("Project does not exist")
        if not kwargs['test']:
            sys.exit(1)
    except ProjectNameAlreadyExists:
        project = Project.remove_project(name)
        print("Successfully removed cobra project {} from the project file".format(name))

    # load config
    Config.getInstance().load_config()

    # github user object
    user = get_github_user()

    if remove_all:
        repo = True
        conda = True
        folder = True

    if repo:
        try:
            repo = user.get_repo(project.repo_name)
            repo.delete()
            print("Successfully removed repo")
        except UnknownObjectException as e:
            if e.status == 404:
                print("Repo could not be deleted! Not existing")
            if e.status == 403:
                print('Permission problem, please reinstall cobra (--clean; --install)')

    if folder:
        delete_path(project.project_path)
        print("Successfully removed folders")

    if conda:
        if check_env_exists(project.conda_name):
            try:
                # remove env
                subprocess.check_output(['conda', 'env', 'remove', '-y', '--name', project.conda_name])
                print("Successfully removed conda environment")
            except subprocess.CalledProcessError:
                print('Environment was not deactivated!')
                print('Remove the environment by hand with the following command:\n')
                print('conda deactivate\n')
                print('conda env remove --name ' + project.conda_name + '-y')
                print('\n')
        else:
            print("Conda environment does not exist")


def entry_point():
    main()


if __name__ == '__main__':
    entry_point()
